@@ -1,62 +1,24 @@
-/* vi: set sw=4 ts=4: */
-/*
- * Copyright (C) 2001 Christian Hohnstaedt.
+/* vi: set sw=4 ts=4:
  *
- *  All rights reserved.
+ * Copyright (C) 2001 - 2009 Christian Hohnstaedt.
  *
- *
- *  Redistribution and use in source and binary forms, with or without 
- *  modification, are permitted provided that the following conditions are met:
- *
- *  - Redistributions of source code must retain the above copyright notice,
- *    this list of conditions and the following disclaimer.
- *  - Redistributions in binary form must reproduce the above copyright notice,
- *    this list of conditions and the following disclaimer in the documentation
- *    and/or other materials provided with the distribution.
- *  - Neither the name of the author nor the names of its contributors may be 
- *    used to endorse or promote products derived from this software without
- *    specific prior written permission.
- *
- *
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
- * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
- * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
- * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
- * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
- * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
- * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
- * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
- * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
- * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
- * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
- *
- *
- * This program links to software with different licenses from:
- *
- *	http://www.openssl.org which includes cryptographic software
- * 	written by Eric Young (eay@cryptsoft.com)"
- *
- *	http://www.sleepycat.com
- *
- *	http://www.trolltech.com
- * 
- *
- *
- * http://www.hohnstaedt.de/xca
- * email: christian@hohnstaedt.de
- *
- * $Id: pki_x509.cpp,v 1.91 2004/02/18 00:03:06 chris2511 Exp $
- *
- */                           
-
+ * All rights reserved.
+ */
 
 
 #include "pki_x509.h"
+#include "pki_evp.h"
+#include "pki_scard.h"
 #include "func.h"
+#include "base.h"
+#include "exception.h"
+#include "pass_info.h"
+#include "widgets/MainWindow.h"
+#include <qdir.h>
 
-QPixmap *pki_x509::icon[4] = { NULL, NULL, NULL, NULL };
+QPixmap *pki_x509::icon[5] = { NULL, NULL, NULL, NULL, NULL };
 
-pki_x509::pki_x509(X509 *c) 
+pki_x509::pki_x509(X509 *c)
 	:pki_x509super()
 {
 	init();
@@ -64,7 +26,7 @@ pki_x509::pki_x509(X509 *c)
 	openssl_error();
 }
 
-pki_x509::pki_x509(const pki_x509 *crt) 
+pki_x509::pki_x509(const pki_x509 *crt)
 	:pki_x509super(crt->desc)
 {
 	init();
@@ -78,7 +40,7 @@ pki_x509::pki_x509(const pki_x509 *crt)
 	caSerial = crt->caSerial;
 	caTemplate = crt->caTemplate;
 	crlDays = crt->crlDays;
-	lastCrl = crt->lastCrl;
+	crlExpiry = crt->crlExpiry;
 	isrevoked = isrevoked;
 	openssl_error();
 }
@@ -92,26 +54,42 @@ pki_x509::pki_x509(const QString name)
 	openssl_error();
 }
 
-void pki_x509::fload(const QString fname)
+void pki_x509::fromPEM_BIO(BIO *bio, QString name)
 {
-	FILE *fp = fopen(fname.latin1(),"r");
 	X509 *_cert;
-	if (fp != NULL) {
-		_cert = PEM_read_X509(fp, NULL, NULL, NULL);
-		if (!_cert) {
-			ign_openssl_error();
-			rewind(fp);
-			_cert = d2i_X509_fp(fp, NULL);
-		}
-		openssl_error();
-		autoIntName();
-		if (getIntName().isEmpty())
-			setIntName(rmslashdot(fname));
-	}	
-	else fopen_error(fname);
-	fclose(fp);
+	_cert = PEM_read_bio_X509(bio, NULL, NULL, NULL);
+	openssl_error(name);
 	X509_free(cert);
 	cert = _cert;
+	autoIntName();
+	if (getIntName().isEmpty())
+		setIntName(rmslashdot(name));
+	trust = 1;
+	efftrust = 1;
+}
+
+void pki_x509::fload(const QString fname)
+{
+	FILE *fp = fopen(QString2filename(fname), "r");
+	X509 *_cert;
+	if (!fp) {
+		fopen_error(fname);
+		return;
+	}
+	_cert = PEM_read_X509(fp, NULL, NULL, NULL);
+	if (!_cert) {
+		ign_openssl_error();
+		rewind(fp);
+		_cert = d2i_X509_fp(fp, NULL);
+	}
+	fclose(fp);
+	openssl_error(fname);
+
+	X509_free(cert);
+	cert = _cert;
+	autoIntName();
+	if (getIntName().isEmpty())
+			setIntName(rmslashdot(fname));
 	trust = 1;
 	efftrust = 1;
 }
@@ -133,10 +111,13 @@ void pki_x509::init()
 	caSerial = 1;
 	caTemplate = "";
 	crlDays = 30;
-	lastCrl.now();
+	crlExpiry.now();
 	class_name = "pki_x509";
 	cert = NULL;
 	isrevoked = false;
+	dataVersion=1;
+	pkiType=x509;
+	cols=6;
 }
 
 void pki_x509::setSerial(const a1int &serial)
@@ -154,26 +135,142 @@ a1int pki_x509::getSerial() const
 	return a;
 }
 
+a1int pki_x509::hashInfo(const EVP_MD *md) const
+{
+	unsigned char digest[EVP_MAX_MD_SIZE];
+	unsigned len = 0;
+	if (!ASN1_item_digest(ASN1_ITEM_rptr(X509_CINF), md,
+				(char*)cert->cert_info,digest,&len))
+		openssl_error();
+	a1int a;
+	a.setRaw(digest,len);
+	return a;
+}
+
+a1int pki_x509::getQASerial(const a1int &secret) const
+{
+	ASN1_INTEGER *hold = cert->cert_info->serialNumber;
+	cert->cert_info->serialNumber = secret.get();
+	a1int ret = hashInfo(EVP_md5());
+	ASN1_INTEGER_free(cert->cert_info->serialNumber);
+	cert->cert_info->serialNumber = hold;
+	return ret;
+}
+
+void pki_x509::load_token(pkcs11 &p11, CK_OBJECT_HANDLE object)
+{
+	QStringList sl = p11.tokenInfo();
+	const unsigned char *p;
+	unsigned long s;
+	QString desc;
+
+	pk11_attr_ulong type(CKA_CERTIFICATE_TYPE);
+	p11.loadAttribute(type, object);
+	if (type.getValue() != CKC_X_509)
+		throw errorEx(QString("Unsupported Certificate type %1"
+			).arg(type.getValue()));
+
+	try {
+		pk11_attr_data label(CKA_LABEL);
+		p11.loadAttribute(label, object);
+		desc = label.getText();
+	} catch(errorEx &err) {
+		printf("No Cert Label: %s\n", err.getCString());
+		// IGNORE
+	}
+	pk11_attr_data x509(CKA_VALUE);
+	p11.loadAttribute(x509, object);
+	s = x509.getValue(&p);
+	if (cert)
+		X509_free(cert);
+	cert = d2i_X509(NULL, &p, s);
+
+	if (desc.isEmpty()) {
+		try {
+			x509name xn;
+
+			pk11_attr_data subj(CKA_SUBJECT);
+			p11.loadAttribute(subj, object);
+			s = subj.getValue(&p);
+			xn.d2i(p, s);
+			desc = xn.getMostPopular();
+			openssl_error();
+		} catch(errorEx &err) {
+			printf("No Cert Subject: %s\n", err.getCString());
+			// IGNORE
+		}
+	}
+	setIntName(sl[0] + " (" + desc + ")");
+	openssl_error();
+}
+
+void pki_x509::store_token()
+{
+	pki_scard *card = (pki_scard *)privkey;
+	int slot, size, id_size;
+	unsigned char *p, *p1, *id;
+	QList<CK_OBJECT_HANDLE> objects;
+
+	if (!privkey || !privkey->isScard())
+		throw errorEx(tr("No associated Smart card"));
+
+	slot = card->prepare_card();
+
+	size = i2d_X509(cert, NULL);
+	openssl_error();
+	p = p1 = (unsigned char*)OPENSSL_malloc(size);
+	i2d_X509(cert, &p1);
+	openssl_error();
+
+	id_size = card->getIdBin(&id);
+	openssl_error();
+
+	pk11_attlist p11_atts;
+	p11_atts <<
+		pk11_attr_ulong(CKA_CLASS, CKO_CERTIFICATE) <<
+		pk11_attr_ulong(CKA_CERTIFICATE_TYPE, CKC_X_509) <<
+		pk11_attr_bool(CKA_TOKEN, true) <<
+		pk11_attr_data(CKA_VALUE, p, size) <<
+		pk11_attr_data(CKA_ID, id, id_size) <<
+		pk11_attr_data(CKA_LABEL, desc.toUtf8());
+
+	free(p);
+	free(id);
+
+	pkcs11 p11;
+	p11.startSession(slot, true);
+
+	if (card->scardLogin(p11, false).isNull())
+		return;
+
+	p11.createObject(p11_atts);
+}
+
+bool pki_x509::verifyQASerial(const a1int &secret) const
+{
+	return getQASerial(secret) == getSerial();
+}
+
+void pki_x509::set_date(ASN1_TIME **a, const a1time &a1)
+{
+	if (*a != NULL )
+		ASN1_TIME_free(*a);
+
+	*a = a1.get_utc();
+	if (*a == NULL)
+		*a = a1.get();
+
+	openssl_error();
+}
+
 void pki_x509::setNotBefore(const a1time &a1)
 {
-	if (X509_get_notBefore(cert) != NULL ) {
-		ASN1_TIME_free(X509_get_notBefore(cert));
-	}
-	X509_get_notBefore(cert) = a1.get_utc();
-	if (X509_get_notBefore(cert) == NULL)
-		X509_get_notBefore(cert) = a1.get();
-	openssl_error();
+	set_date(&X509_get_notBefore(cert), a1);
 }
 
 void pki_x509::setNotAfter(const a1time &a1)
 {
-	if (X509_get_notAfter(cert) != NULL ) {
-		ASN1_TIME_free(X509_get_notAfter(cert));
-	}
-	X509_get_notAfter(cert) = a1.get_utc();
-	if (X509_get_notAfter(cert) == NULL)
-		X509_get_notAfter(cert) = a1.get();
-	openssl_error();
+	set_date(&X509_get_notAfter(cert), a1);
 }
 
 a1time pki_x509::getNotBefore() const
@@ -217,7 +314,7 @@ void pki_x509::setIssuer(const x509name &n)
 }
 
 void pki_x509::addV3ext(const x509v3ext &e)
-{	
+{
 	if (!e.isValid()) return;
 	X509_EXTENSION *ext = e.get();
 	X509_add_ext(cert, ext, -1);
@@ -225,161 +322,135 @@ void pki_x509::addV3ext(const x509v3ext &e)
 	openssl_error();
 }
 
+void pki_x509::delSigner(pki_base *s)
+{
+	if (s == psigner)
+		psigner = NULL;
+}
+
 bool pki_x509::canSign()
 {
 	BASIC_CONSTRAINTS *bc;
 	int crit;
-	if (!privkey || privkey->isPubKey()) return false;
+	if (!privkey || privkey->isPubKey())
+		return false;
+	if (privkey->isScard() && !pkcs11::loaded())
+		return false;
 	bc = (BASIC_CONSTRAINTS *)X509_get_ext_d2i(cert, NID_basic_constraints, &crit, NULL);
 	openssl_error();
-	if (!bc || !bc->ca) return false;
+	if (!bc || !bc->ca)
+		return false;
 	return true;
 }
-	
-bool pki_x509::hasSubAltName()
+
+bool pki_x509::hasExtension(int nid)
 {
-	STACK_OF(GENERAL_NAME) *subAlt;
-	int crit;
-	subAlt = (STACK_OF(GENERAL_NAME) *)X509_get_ext_d2i(cert, NID_subject_alt_name, &crit, NULL);
-	openssl_error();
-	if (sk_GENERAL_NAME_num(subAlt) < 1) return false;	
-	return true;
+	int i;
+	STACK_OF(X509_EXTENSION) *x = cert->cert_info->extensions;
+
+	for(i = 0; i < sk_X509_EXTENSION_num(x); i++) {
+		X509_EXTENSION *ex = sk_X509_EXTENSION_value(x, i);
+		if (OBJ_obj2nid(ex->object) == nid) {
+			return true;
+		}
+	}
+	return false;
 }
-	
+
 void pki_x509::sign(pki_key *signkey, const EVP_MD *digest)
 {
+	EVP_PKEY *tkey;
 	if (!signkey) {
-		openssl_error("There is no key for signing !");
+		my_error(tr("There is no key for signing !"));
 	}
-	X509_sign(cert, signkey->key, digest);
+	tkey = signkey->decryptKey();
+	openssl_error();
+	X509_sign(cert, tkey, digest);
+	openssl_error();
+	EVP_PKEY_free(tkey);
 	openssl_error();
 }
 
-
-
-/* Save the Certificate to data and back:
- * Version 1:
- * 	int Version
- * 	int size of cert
- * 	cert
- * 	int trust
- * 	int size of revTime
- * 	revocationtime
- */
-
-	
-void pki_x509::fromData(unsigned char *p, int size)
+void pki_x509::fromData(const unsigned char *p, db_header_t *head)
 {
-	int version, sCert, sRev, sLastCrl;
-	unsigned char *p1 = p;
+	int version, size;
+	const unsigned char *p1 = p;
+
 	X509 *cert_sik = cert;
-	version = intFromData(&p1);
-	if (version >=1 || version <= 4) {
-		sCert = intFromData(&p1);
-		cert = d2i_X509(NULL, &p1, sCert);
-		trust = intFromData(&p1);
-		sRev = intFromData(&p1);
-		if (sRev) {
-	//		ASN1_TIME *time;
-			if (version != 3) isrevoked = true;
-			p1 = revoked.d2i(p1, sRev);
-	//		time = d2i_ASN1_TIME(NULL, &p1, sRev);
-	//		revoked = time;
-	//		ASN1_TIME_free(time);
-		}
-		else {
-		   isrevoked = false;
-		   revoked.now();
-		}
-		
-		if (version == 1) {
-			caTemplate="";
-			caSerial=1;
-			lastCrl = NULL;
-			crlDays=30;
-		}
-		
-		if (version >= 2 ) {
-			caSerial = intFromData(&p1);
-			caTemplate = stringFromData(&p1);
-		}
-		if (version >= 3 ) {
-			crlDays = intFromData(&p1);
-			sLastCrl = intFromData(&p1);
-			if (sLastCrl) {
-			   lastCrl.d2i(p1, sLastCrl);
-			}
-		}
-		// version 4 saves a NULL as revoked
-		// version 3 did save a recent date :-((
-	}
-	else { // old version
-		cert = d2i_X509(NULL, &p, size);
-		revoked = NULL;
-		trust = 1;
-		efftrust = 1;
-	}	
+	version = head->version;
+	size = head->len - sizeof(db_header_t);
+
+	cert = D2I_CLASH(d2i_X509, NULL, &p1, size);
+	trust = db::intFromData(&p1);
+	isrevoked = db::boolFromData(&p1);
+	p1 = revoked.d2i(p1, size - (p1-p));
+	caSerial.setHex(db::stringFromData(&p1));
+	caTemplate = db::stringFromData(&p1);
+	crlDays = db::intFromData(&p1);
+	crlExpiry.d2i(p1, size - (p1-p));
+
 	if (cert)
 		X509_free(cert_sik);
 	else
 		cert = cert_sik;
 	openssl_error();
-		
 }
 
 
 unsigned char *pki_x509::toData(int *size)
 {
-#define PKI_DB_VERSION (int)4
 	unsigned char *p, *p1;
-	int sCert = i2d_X509(cert, NULL);
-	int sRev = revoked.derSize();
-	int sLastCrl = lastCrl.derSize();
-	if (!isrevoked) sRev = 0;
-	// calculate the needed size 
-	*size = caTemplate.length() + 1 + sCert + sRev + sLastCrl + (7 * sizeof(int));
+
+	// calculate the needed size
+	*size = i2d_X509(cert, NULL) + 11 + caSerial.toHex().length() +
+		caTemplate.length() + crlExpiry.derSize() + revoked.derSize();
 	openssl_error();
-	p = (unsigned char*)OPENSSL_malloc(*size);
-	p1 = p;
-	intToData(&p1, (PKI_DB_VERSION)); // version
-	intToData(&p1, sCert); // sizeof(cert)
+	p = p1 = (unsigned char*)OPENSSL_malloc(*size);
+
 	i2d_X509(cert, &p1); // cert
-	intToData(&p1, trust); // trust
-	intToData(&p1, sRev); // sizeof(revoked)
-	if (sRev) {
-		p1 = revoked.i2d(p1); // revokation date
-	}
-			 
-	// version 2
-	intToData(&p1, caSerial.getLong()); // the serial if this is a CA
-	stringToData(&p1, caTemplate); // the name of the template to use for signing
+	db::intToData(&p1, trust); // trust
+	db::boolToData(&p1, isrevoked);
+	p1 = revoked.i2d(p1); // revokation date
+
+	// the serial if this is a CA
+	db::stringToData(&p1, caSerial.toHex());
+	// the name of the template to use for signing
+	db::stringToData(&p1, caTemplate);
 	// version 3
-	intToData(&p1, crlDays); // the CRL period
-	intToData(&p1, sLastCrl); // size of last CRL
-	if (sLastCrl) {
-		p1 = lastCrl.i2d(p1); // last CRL date
-	}
+	db::intToData(&p1, crlDays); // the CRL period
+	p1 = crlExpiry.i2d(p1); // last CRL date
 	openssl_error();
+	if (*size != p1-p) {
+		printf("pki_x509::toData: Size = %d, real size=%zd\n", *size, p1-p);
+		//abort();
+	}
 	return p;
 }
 
+void pki_x509::writeDefault(const QString fname)
+{
+	writeCert(fname + QDir::separator() + getIntName() + ".crt",
+			true, false);
+}
+
 void pki_x509::writeCert(const QString fname, bool PEM, bool append)
 {
 	FILE *fp;
-	char *_a = "a", *_w="w", *p = _w;
-	if (append) p=_a;
-	fp = fopen(fname.latin1(), p);
+	const char *p = "w";
+	if (append)
+		p = "a";
+	fp = fopen(QString2filename(fname), p);
 	if (fp != NULL) {
 		if (cert){
-			if (PEM) 
+			if (PEM)
 				PEM_write_X509(fp, cert);
 			else
 				i2d_X509_fp(fp, cert);
-			openssl_error();
 		}
-	}
-	else fopen_error(fname);
-	fclose(fp);
+		fclose(fp);
+		openssl_error();
+	} else
+		fopen_error(fname);
 }
 
 bool pki_x509::compare(pki_base *ref)
@@ -394,23 +465,23 @@ bool pki_x509::cmpIssuerAndSerial(pki_x509 *refcert)
 	bool ret =  X509_issuer_and_serial_cmp(cert, refcert->cert);
 	openssl_error();
 	return ret;
-			  
-}	
-	
+
+}
+
 bool pki_x509::verify(pki_x509 *signer)
 {
-	if (psigner == signer) return true;
-	if ((psigner != NULL )||( signer == NULL)) return false;
+	if (psigner == signer)
+		return true;
+	if ((psigner != NULL )||( signer == NULL))
+		return false;
 	X509_NAME *subject =  X509_get_subject_name(signer->cert);
 	X509_NAME *issuer = X509_get_issuer_name(cert);
 	openssl_error();
 	if (X509_NAME_cmp(subject, issuer)) {
 		return false;
 	}
-	pki_key *pkey = signer->getPubKey();
-	int i = X509_verify(cert,pkey->key);
+	int i = X509_verify(cert, X509_get_pubkey(signer->cert));
 	ign_openssl_error();
-	if (pkey) delete(pkey);
 	if (i>0) {
 		psigner = signer;
 		return true;
@@ -423,13 +494,13 @@ pki_key *pki_x509::getPubKey() const
 {
 	EVP_PKEY *pkey = X509_get_pubkey(cert);
 	openssl_error();
-	pki_key *key = new pki_key(pkey);	
+	pki_evp *key = new pki_evp(pkey);
 	return key;
 }
 
 void pki_x509::setPubKey(pki_key *key)
 {
-	 X509_set_pubkey(cert, key->getKey());
+	 X509_set_pubkey(cert, key->getPubKey());
 }
 
 QString pki_x509::fingerprint(const EVP_MD *digest)
@@ -475,37 +546,28 @@ int pki_x509::resetTimes(pki_x509 *signer)
 	openssl_error();
 	return ret;
 }
-	
 
-pki_x509 *pki_x509::getSigner() { return (psigner); }
-
-void pki_x509::delSigner(pki_x509 *s) 
-{
-	if (s == psigner) 
-		psigner = NULL;
-}
-
-QString pki_x509::printV3ext()
+extList pki_x509::getV3ext()
 {
 	extList el;
 	el.setStack(cert->cert_info->extensions);
-	QString text = el.getHtml("<br>");
-	openssl_error();
-	return text;
+	return el;
 }
 
 x509v3ext pki_x509::getExtByNid(int nid)
 {
-	extList el;
-	x509v3ext e;
-	el.setStack(cert->cert_info->extensions);
-	
-	for (unsigned int i=0; i< el.count(); i++){
-		if (el[i].nid() == nid) return el[i];
-	}			
-	return e;
-}	
+	extList el = getV3ext();
+	int i = el.idxByNid(nid);
 
+	if (i == -1)
+		return x509v3ext();
+	return el[i];
+}
+
+pki_x509 *pki_x509::getSigner()
+{
+	return (pki_x509 *)psigner;
+}
 
 int pki_x509::getTrust()
 {
@@ -533,7 +595,6 @@ void pki_x509::setEffTrust(int t)
 		efftrust = t;
 }
 
-
 bool pki_x509::isRevoked()
 {
 	return isrevoked ;
@@ -555,14 +616,14 @@ a1time &pki_x509::getRevoked()
 {
 	return revoked;
 }
-	
+
 void pki_x509::setRevoked(const a1time &when)
 {
 	isrevoked = true;
 	revoked = when;
 	setEffTrust(0);
 	setTrust(0);
-	openssl_error();	
+	openssl_error();
 }
 
 int pki_x509::calcEffTrust()
@@ -579,13 +640,11 @@ int pki_x509::calcEffTrust()
 	}
 	//we must look at the parent certs
 	pki_x509 *signer = getSigner();
-	pki_x509 *prevsigner = this;
-	while (mytrust==1 && signer != NULL && signer != prevsigner) {
+	while (mytrust == 1 && signer && signer != this) {
 		mytrust = signer->getTrust();
-		prevsigner = signer;
 		signer = signer->getSigner();
 	}
-	
+
 	if (mytrust == 1) mytrust = 0;
 	efftrust = mytrust;
 	return mytrust;
@@ -605,41 +664,12 @@ QString pki_x509::getTemplate(){ return caTemplate; }
 
 void pki_x509::setTemplate(QString s) {if (s.length()>0) caTemplate = s; }
 
-void pki_x509::setLastCrl(const a1time &time)
+void pki_x509::setCrlExpiry(const a1time &time)
 {
-	lastCrl = time;
+	crlExpiry = time;
 	openssl_error();
 }
 
-QString pki_x509::tinyCAfname()
-{
-	QString col;
-	x509name x = getSubject();
-	col = x.getEntryByNid(NID_commonName) 
-	    +(x.getEntryByNid(NID_commonName) == "" ? " :" : ":")
-	    + x.getEntryByNid(NID_pkcs9_emailAddress) 
-	    +(x.getEntryByNid(NID_pkcs9_emailAddress) == "" ? " :" : ":")
-	    + x.getEntryByNid(NID_organizationalUnitName) 
-	    +(x.getEntryByNid(NID_organizationalUnitName) == "" ? " :" : ":")
-	    + x.getEntryByNid(NID_organizationName) 
-	    +(x.getEntryByNid(NID_organizationName) == "" ? " :" : ":")
-	    + x.getEntryByNid(NID_localityName) 
-	    +(x.getEntryByNid(NID_localityName) == "" ? " :" : ":")
-	    + x.getEntryByNid(NID_stateOrProvinceName) 
-	    +(x.getEntryByNid(NID_stateOrProvinceName) == "" ? " :" : ":")
-	    + x.getEntryByNid(NID_countryName) 
-	    +(x.getEntryByNid(NID_countryName) == "" ? " :" : ":");
-
-	int len = col.length();
-	unsigned char *buf = (unsigned char *)OPENSSL_malloc(len * 2 + 3);
-	
-	EVP_EncodeBlock(buf, (unsigned char *)col.latin1(), len );
-	col = (char *)buf;
-	OPENSSL_free(buf);
-	col += ".pem";
-	return col;
-}
-
 x509rev pki_x509::getRev()
 {
 	x509rev a;
@@ -647,28 +677,46 @@ x509rev pki_x509::getRev()
 	a.setSerial(getSerial());
 	return a;
 }
-	
-void pki_x509::updateView()
+
+QVariant pki_x509::column_data(int col)
 {
-	pki_base::updateView();
-	QString truststatus[] = 
+	QString truststatus[] =
 		{ tr("Not trusted"), tr("Trust inherited"), tr("Always Trusted") };
+
+	switch (col) {
+		case 0:
+			return QVariant(getIntName());
+		case 1:
+			return QVariant(getSubject().getEntryByNid(NID_commonName));
+		case 2:
+			return QVariant(getSerial().toHex());
+		case 3:
+			return QVariant(getNotAfter().toSortable());
+		case 4:
+			return QVariant(truststatus[ getTrust() ]);
+		case 5:
+			if (isRevoked())
+				return QVariant(getRevoked().toSortable());
+			else if (canSign())
+				return QVariant(tr("CRL expires: ")+ crlExpiry.toSortable());
+	}
+	return QVariant();
+
+}
+
+QVariant pki_x509::getIcon(int column)
+{
 	int pixnum = 0;
-	if (!pointer) return;
+	if (column != 0)
+		return QVariant();
+
 	if (getRefKey()) {
 		pixnum += 1;
 	}
-	if (calcEffTrust() == 0){ 
+	if (calcEffTrust() == 0){
 		pixnum += 2;
-	}	
-	pointer->setPixmap(0, *icon[pixnum]);
-	pointer->setText(0, getIntName());
-	pointer->setText(1, getSubject().getEntryByNid(NID_commonName));
-	pointer->setText(2, getSerial().toHex() );  
-	pointer->setText(3, getNotAfter().toSortable() );  
-	pointer->setText(4, truststatus[ getTrust() ]);  
-	if (isRevoked())
-		pointer->setText(5, getRevoked().toSortable());
+	}
+	return QVariant(*icon[pixnum]);
 }
 
 QString pki_x509::getSigAlg()
@@ -682,3 +730,68 @@ const EVP_MD *pki_x509::getDigest()
 	return EVP_get_digestbyobj(cert->sig_alg->algorithm);
 }
 
+void pki_x509::oldFromData(unsigned char *p, int size)
+{
+	int version, sCert, sRev, sLastCrl;
+	const unsigned char *p1 = p;
+	X509 *cert_sik = cert;
+	version = intFromData(&p1);
+	if (version >=1 && version <= 5) {
+		sCert = intFromData(&p1);
+		cert = D2I_CLASH(d2i_X509, NULL, &p1, sCert);
+		trust = intFromData(&p1);
+		sRev = intFromData(&p1);
+		if (sRev) {
+			if (version != 3) isrevoked = true;
+			p1 = revoked.d2i(p1, sRev);
+		}
+		else {
+		   isrevoked = false;
+		   revoked.now();
+		}
+
+		if (version == 1) {
+			caTemplate="";
+			caSerial=1;
+			crlExpiry.now();
+			crlDays=30;
+		}
+
+		if (version >= 2 ) {
+			if (version >= 5)
+				caSerial.setHex(db::stringFromData(&p1));
+			else {
+				int i = intFromData(&p1);
+				if (i>=0)
+					caSerial = i;
+				else {
+					caSerial = getSerial();
+					++caSerial;
+				}
+			}
+			caTemplate = db::stringFromData(&p1);
+		}
+		if (version >= 3 ) {
+			crlDays = intFromData(&p1);
+			sLastCrl = intFromData(&p1);
+			if (sLastCrl) {
+			   crlExpiry.d2i(p1, sLastCrl);
+			}
+		}
+		// version 4 saves a NULL as revoked
+		// version 3 did save a recent date :-((
+	}
+	else { // old version
+		p1 = p;
+		cert = D2I_CLASH(d2i_X509, NULL, &p1, size);
+		revoked = NULL;
+		trust = 1;
+		efftrust = 1;
+	}
+	if (cert)
+		X509_free(cert_sik);
+	else
+		cert = cert_sik;
+	openssl_error();
+}
+
